import ntp.packet
import ntp.ntpc
import threading
from contextlib import AbstractContextManager
import time
import argparse
import os
import sys
import pwd
import grp
import logging
import logging.handlers
import pyagentx
import datetime


BASE_OID_ENTERPRISE             = '1.3.6.1.4.1'
BASE_OID_HOTTIS                 = BASE_OID_ENTERPRISE + '.9676'
BASE_OID_HOTTIS_NTPSEC          = BASE_OID_HOTTIS + '.123'

# just the prefix where the objects are below
SYSINFO_PREFIX                  = '1'
SYSSTATS_PREFIX                 = '2'
PEERS_PREFIX                    = '3'
NUMBER_OF_PEERS_PREFIX          = PEERS_PREFIX + '.1'

# this is for a table
# 2 is the prefix
# the first 1 is for the table in the mib
# the second 1 is for the entries in the table in the mib
TABLE_OF_PEERS_PREFIX           = PEERS_PREFIX + '.2.1'


def int_scale1k(x):
    return int (x * 1000)

def int_scale1M(x):
    return int (x * 1000000)

def pass_value(x):
    return x

def string_ntp_seconds(x):
    return ntp.ntpc.prettydate(x).split(' ')[1]


SYSINFO_KEYS = [ 
    ['peeradr', pyagentx.TYPE_OCTETSTRING, pass_value],
    ['peermode', pyagentx.TYPE_INTEGER, pass_value],
    ['leap', pyagentx.TYPE_INTEGER, pass_value],
    ['stratum', pyagentx.TYPE_INTEGER, pass_value],
    ['precision', pyagentx.TYPE_INTEGER, pass_value], 
    ['rootdelay', pyagentx.TYPE_INTEGER, int_scale1M],
    ['rootdisp', pyagentx.TYPE_INTEGER, int_scale1M],
    ['rootdist', pyagentx.TYPE_INTEGER, int_scale1M],
    ['refid', pyagentx.TYPE_OCTETSTRING, pass_value],
    ['reftime', pyagentx.TYPE_OCTETSTRING, string_ntp_seconds],
    ['sys_jitter', pyagentx.TYPE_INTEGER, int_scale1M],
    ['clk_jitter', pyagentx.TYPE_INTEGER, int_scale1M],
    ['clk_wander', pyagentx.TYPE_INTEGER, int_scale1M],
]

SYSSTATS_KEYS = [
    ['ss_uptime', pyagentx.TYPE_INTEGER, pass_value ],
    ['ss_numctlreq', pyagentx.TYPE_INTEGER, pass_value ],
    ['ss_reset', pyagentx.TYPE_INTEGER, pass_value ],
    ['ss_received', pyagentx.TYPE_COUNTER64, pass_value ],
    ['ss_badformat', pyagentx.TYPE_COUNTER64, pass_value ],
    ['ss_declined', pyagentx.TYPE_COUNTER64, pass_value ],
    ['ss_restricted', pyagentx.TYPE_COUNTER64, pass_value ],
    ['ss_limited', pyagentx.TYPE_COUNTER64, pass_value ],
    ['ss_kodsent', pyagentx.TYPE_COUNTER64, pass_value ],
    ['ss_processed', pyagentx.TYPE_COUNTER64, pass_value ]
]

PEER_KEYS = [
    ['srchost', pyagentx.TYPE_OCTETSTRING, pass_value],
    ['srcadr', pyagentx.TYPE_OCTETSTRING, pass_value],
    ['refid', pyagentx.TYPE_OCTETSTRING, pass_value],
    ['stratum', pyagentx.TYPE_INTEGER, pass_value],
    ['hmode', pyagentx.TYPE_INTEGER, pass_value],
    ['ppoll', pyagentx.TYPE_INTEGER, pass_value],
    ['hpoll', pyagentx.TYPE_INTEGER, pass_value],
    ['rec', pyagentx.TYPE_OCTETSTRING, string_ntp_seconds],
    ['reach', pyagentx.TYPE_INTEGER, pass_value],
    ['delay', pyagentx.TYPE_INTEGER, int_scale1k],
    ['offset', pyagentx.TYPE_INTEGER, int_scale1M],
    ['jitter', pyagentx.TYPE_INTEGER, int_scale1M],
]




class DataStore(AbstractContextManager):
    def __init__(self):
        self.lock = threading.Lock()
        self.data = {}

    def update_data(self, data):
        self.data.clear()
        self.data.update(data)

    def __enter__(self):
        self.lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.lock.release()


class NtpDataCollector(threading.Thread):
    def __init__(self, ntpserver='localhost', period=60):
        self.ntpserver = ntpserver
        self.period = period
        self.stop_event = threading.Event()

        threading.Thread.__init__(self)

    def run(self):
        while not self.stop_event.is_set():
            try:
                logger.debug('Query ntp server')
                tmp_data_store = {}

                session = ntp.packet.ControlSession()
                session.openhost(self.ntpserver)

                sysinfo_vars = session.readvar(0, [ x[0] for x in SYSINFO_KEYS ])
                logger.debug(f"{sysinfo_vars=}")
                tmp_data_store['sysinfo'] = dict(sysinfo_vars)

                sysstats_vars = session.readvar(0, [ x[0] for x in SYSSTATS_KEYS ])
                logger.debug(f"{sysstats_vars=}")
                tmp_data_store['sysstats'] = dict(sysstats_vars)

                peers = session.readstat()
                tmp_data_store['peers'] = {}
                for peer in peers:
                    peer_vars = session.readvar(peer.associd, [ x[0] for x in PEER_KEYS ])
                    tmp_data_store['peers'][peer.associd] = dict(peer_vars)
                    logger.debug(f"{peer.associd=}, {peer_vars=}")

                with globalDataStore as ds:
                    ds.update_data(tmp_data_store)
            except ntp.packet.ControlException as e:
                logger.error(f"ntp.packet.ControlException while querying NTP server: {str(e)}")
            finally:
                time.sleep(self.period)

        logger.info('NtpDataCollector terminating')

    def stop(self):
        self.stop_event.set()
        self.join()


class NtpsecDataUpdater(pyagentx.Updater):
    def update(self):
        with globalDataStore as ds:
            if ds.data:
                try:
                    logger.debug("Updating data store")                    
                    for index, data_spec in enumerate(SYSINFO_KEYS, start=1):
                        # logger.debug(f"local: {index=} {data_spec=}")
                        try:
                            oid_prefix = f"{SYSINFO_PREFIX}.{index}"
                            self._data[oid_prefix] = {
                                'name': oid_prefix, 
                                'type': data_spec[1], 
                                'value': data_spec[2](ds.data['sysinfo'][data_spec[0]])
                            }
                        except KeyError as e:
                            logger.error(f"key {data_spec[0]} missing in local, skip it: {str(e)}")

                    for index, data_spec in enumerate(SYSSTATS_KEYS, start=1):
                        # logger.debug(f"local: {index=} {data_spec=}")
                        try:
                            oid_prefix = f"{SYSSTATS_PREFIX}.{index}"
                            self._data[oid_prefix] = {
                                'name': oid_prefix, 
                                'type': data_spec[1], 
                                'value': data_spec[2](ds.data['sysstats'][data_spec[0]])
                            }
                        except KeyError as e:
                            logger.error(f"key {data_spec[0]} missing in local, skip it: {str(e)}")

                    number_of_peers = len(ds.data['peers'])
                    # logger.debug(f"number of peers: {number_of_peers}")
                    number_of_peers_oid_prefix = f"{NUMBER_OF_PEERS_PREFIX}"
                    self._data[number_of_peers_oid_prefix] = {
                        'name': number_of_peers_oid_prefix, 
                        'type': pyagentx.TYPE_INTEGER,
                        'value': number_of_peers
                    }
                    for peer_index, (associd, peer) in enumerate(ds.data['peers'].items(), start=1):
                        # logger.debug(f"peer: {peer}")
                        index_oid_prefix = f"{TABLE_OF_PEERS_PREFIX}.1.{peer_index}"
                        self._data[index_oid_prefix] = {
                            'name': index_oid_prefix, 
                            'type': pyagentx.TYPE_INTEGER,
                            'value': peer_index
                        }
                        associd_oid_prefix = f"{TABLE_OF_PEERS_PREFIX}.2.{peer_index}"
                        self._data[associd_oid_prefix] = {
                            'name': associd_oid_prefix, 
                            'type': pyagentx.TYPE_INTEGER,
                            'value': associd
                        }
                        for key_index, data_spec in enumerate(PEER_KEYS, start=3):
                            # logger.debug(f"peer: {associd=} {key_index=} {data_spec=}")
                            try:
                                oid_prefix = f"{TABLE_OF_PEERS_PREFIX}.{key_index}.{peer_index}"
                                self._data[oid_prefix] = {
                                    'name': oid_prefix, 
                                    'type': data_spec[1], 
                                    'value': data_spec[2](peer[data_spec[0]])
                                }
                            except KeyError as e:
                                logger.error(f"key {data_spec[0]} missing in peer {associd}, skip it: {str(e)}")
                except Exception as e:
                    logger.error(f"Failed to update: {type(e)} {e}")



class NtpsecAgent(pyagentx.Agent):
    def __init__(self, period=1, agent_id='NtpsecAgent', socket_path=None):
        logger.info('Agent created')
        self.period = period
        super().__init__()

    def setup(self):
        logger.info('Agent setup')
        self.register(BASE_OID_HOTTIS_NTPSEC, NtpsecDataUpdater, freq=self.period)




def daemonize(pid_filename):
    if os.fork() > 0:
        sys.exit(0)

    os.setsid()
    pid = os.fork()
    if pid > 0:
        with open(pid_filename, 'w') as pid_file:
            pid_file.write(str(pid))
        sys.exit(0)

    sys.stdout.flush()
    sys.stderr.flush()

    with open("/dev/null", "r") as devnull:
        os.dup2(devnull.fileno(), sys.stdin.fileno())
    with open("/tmp/agentx-ntpsec.log", "a+") as log:
        os.dup2(log.fileno(), sys.stdout.fileno())
        os.dup2(log.fileno(), sys.stderr.fileno())


def set_user_group(user, group):
    if group:
        try:
            gid = grp.getgrnam(group).gr_gid
        except KeyError:
            logger.error(f"Group {group} does not exist")
            sys.exit(1)
        os.setgid(gid)
    if user:
        try:
            uid = pwd.getpwnam(user).pw_uid
        except KeyError:
            logger.error(f"user {user} does not exist")
            sys.exit(1)
        os.setuid(uid)

def setup_logging(verbose):
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
            level=log_level,
            format="%(name)s - %(levelname)s - %(message)s",
            handlers=[logging.handlers.SysLogHandler(address='/dev/log')]
    )
    logger = logging.getLogger('agentx-ntpsec')
    stdout_handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    stdout_handler.setFormatter(formatter)
    pyagentx.setup_logging(debug=verbose)
    logger.addHandler(stdout_handler)
    return logger


if __name__ == '__main__':
    pid_filename = '/tmp/agentx-ntpsec.pid'

    parser = argparse.ArgumentParser(description='snmpd agentx extension for ntpsec')
    parser.add_argument('--period', '-p',
                        help='Period to query the NTP server, in seconds, default 60s',
                        type=int,
                        required=False,
                        default=60)
    parser.add_argument('--ntpserver', '-n',
                        help='NTP server to query, default is localhost',
                        required=False,
                        default='localhost')
    parser.add_argument('--daemonize', '-d',
                        help='Run process in background',
                        required=False,
                        action='store_true',
                        default=False)
    parser.add_argument('--pid',
                        help=f"pid-file when running as daemon, default is {pid_filename}",
                        required=False,
                        default=pid_filename)
    parser.add_argument('-u', '--user',
                        help="Set uid of process",
                        required=False,
                        default='')
    parser.add_argument('-g', '--group',
                        help="Set gid of process",
                        required=False,
                        default='')
    parser.add_argument('--verbose', '-v',
                        help='Enable debug output',
                        required=False,
                        action='store_true',
                        default=False)

    args = parser.parse_args()
    
    if args.daemonize:
        daemonize(pid_filename)
    set_user_group(args.user, args.group)
    logger = setup_logging(args.verbose)


    ntpserver = args.ntpserver
    period = args.period

    try:
        globalDataStore = DataStore()
        ndc = NtpDataCollector(ntpserver=ntpserver, period=period)
        ndc.start()

        nsax = NtpsecAgent(period=period)
        nsax.start()
    except Exception as e:
        logger.error(f"Unhandled exception: {e}")
        nsax.stop()
        ndc.stop()
    except KeyboardInterrupt:
        nsax.stop()
        ndc.stop()

